﻿//-------------------------------------------------------------------------------
// アプリケーション名：InputViewer インプットビューア
// 概要：キーボード、マウスの入力状態をデスクトップ画面に表示する
// 
// 作成者：icomato
// 作成日：2022/08/13
// 
// 変更日     変更者  内容
//-------------------------------------------------------------------------------
// 2022/08/13 icomato 新規作成
//-------------------------------------------------------------------------------

using System.Collections.Generic;
using System;
using System.Diagnostics;
using System.Drawing;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.IO;

namespace InputViewer
{
	/// <summary>
	/// メインフォーム
	/// </summary>
	public partial class MainForm : Form
	{
		//----------------------------------------------------------------
		// ＜参考＞
		//	 ●鳩でもわかるC#
		//		キーボードフック
		//		https://lets-csharp.com/keyboard-hook/	 
		//		マウスフック
		//		https://lets-csharp.com/mouse-hook/
		//	 ●KT Software
		//		仮想キーコード
		//		https://kts.sakaiweb.com/virtualkeycodes.html
		//	 ●Microsoft
		//		ウィンドウズメッセージ WM_MOUSEWHEEL
		//		https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-mousewheel
		//		ウィンドウズメッセージ WM_MBUTTONDOWN
		//		https://docs.microsoft.com/ja-jp/windows/win32/inputdev/wm-mbuttondown
		//		Thread.Sleep
		//		https://docs.microsoft.com/ja-jp/dotnet/api/system.threading.thread.sleep?view=net-6.0
		//	 ●DOBON.NET
		//		タスクトレイ
		//		https://dobon.net/vb/dotnet/control/showtasktrayicon.html
		//	 ●アイコンコンバータ
		//		アイコン生成
		//		https://service.tree-web.net/icon_converter/
		//----------------------------------------------------------------

		#region 【Win32定数】仮想キーコード
		public const int VK_LSHIFT = 0xA0;      // 160	左Shiftキー	左[Shift]
		public const int VK_RSHIFT = 0xA1;      // 161	右Shiftキー	右[Shift]
		public const int VK_LCONTROL = 0xA2;        // 162	左Ctrlキー	左[Ctrl]
		public const int VK_RCONTROL = 0xA3;        // 163	右Ctrlキー	右[Ctrl]
		public const int VK_LMENU = 0xA4;       // 164	左Altキー	左[Alt]
		public const int VK_RMENU = 0xA5;       // 165	右Altキー	右[Alt]

		public const int VK_A = 0x41;       // 65	Aキー	[A]
		public const int VK_Z = 0x5A;       // 90	Zキー	[Z]
		#endregion 【Win32定数】仮想キーコード

		#region 【メンバ変数】
		/// <summary>
		/// キーボードフック
		/// </summary>
		private KeyboardHook keyboardHook = new KeyboardHook();
		/// <summary>
		/// マウスフック
		/// </summary>
		private MouseHook mouseHook = new MouseHook();

		/// <summary>
		/// カラー マウス押下色
		/// </summary>
		private Color colDown = Color.Black;
		/// <summary>
		/// カラー マウス押上色
		/// </summary>
		private Color colUp = Color.Transparent;

		/// <summary>
		/// マウス左ボタン 押下状態
		/// </summary>
		private bool mouseLeft = false;
		/// <summary>
		/// マウス右ボタン 押下状態
		/// </summary>
		private bool mouseRight = false;
		/// <summary>
		/// マウス中ボタン 押下状態
		/// </summary>
		private bool mouseMiddle = false;

		/// <summary>
		/// キーボード シフトキー 押下状態
		/// </summary>
		private bool keyShift = false;
		/// <summary>
		/// キーボード Ctrlキー 押下状態
		/// </summary>
		private bool keyCtrl = false;
		/// <summary>
		/// キーボード Altキー 押下状態
		/// </summary>
		private bool keyAlt = false;

		/// <summary>
		/// キーボード 入力履歴状態
		/// </summary>
		private SortedList<DateTime, string> keyHistroies = new SortedList<DateTime, string>();
		/// <summary>
		/// キーボード 入力履歴 処理状態・・・true:処理中、false:処理中ではない
		/// </summary>
		private bool isKeyHistroiesWorking = false;
		#endregion 【メンバ変数】

		#region 【コンストラクタ】
		/// <summary>
		/// コンストラクタ
		/// </summary>
		public MainForm()
		{
			InitializeComponent();

			// ラベル文字列初期化
			this.lblKeyboard.Text = "";
			this.lblKeyHistory.Text = "";
			this.lblMouseText.Text = "";

			//-----------------------
			// コントロール表示設定
			//-----------------------
			this.FormBorderStyle = FormBorderStyle.FixedSingle;
			this.pnlFrame.BackColor = Color.PaleTurquoise;     // パネルをForm非透明色にする
			this.pnlInductionFrame.Visible = true;                  // 右クリックできます 表示


			this.TransparencyKey = Color.White;
		}
		#endregion 【コンストラクタ】

		//----------------------------------
		// イベント
		//----------------------------------

		#region 【イベント】 フォームロードイベント
		/// <summary>
		/// フォームロードイベント
		/// </summary>
		/// <param name="e"></param>
		private void MainForm_Load(object sender, EventArgs e)
		{
			// ＜フォームロードイベント＞以下の初期化処理はFormが読み込み時に処理を行う

			//-----------------------
			// キーボードフック
			//-----------------------
			keyboardHook.KeyEvent += KeyboardHook_KeyEvent;
			keyboardHook.Hook();

			//-----------------------
			// マウスフック
			//-----------------------
			mouseHook.MouseEvent += Hook_MouseMoveEvent;
			mouseHook.Hook();

			//-------------------------------
			// キーボード入力履歴表示タイマ
			//-------------------------------
			this.timerHistoryEraser.Start();

			//-------------------------------
			// タスクトレイ
			//-------------------------------
			//NotifyIconをタスクトレイに表示する
			this.notifyIconApp.Visible = true;

			//-----------------------
			// 設定読み込み
			//-----------------------
			this.LoadSettings();

		}
		#endregion 【イベント】 フォームロードイベント

		#region 【イベント】 キーフック　キー入力イベント
		/// <summary>
		/// キーフック　キー入力イベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void KeyboardHook_KeyEvent(object sender, KeyEventArg e)
		{
			KeysConverter kc = new KeysConverter();
			int keyCode = e.keyCode;
			bool isDown = e.isDown;

			//--------------------------
			// キーボード処理
			//--------------------------
			{
				string keybordText = "";
				bool isKeybordText = false;

				//--------------------------
				// キーボード判定処理
				//--------------------------
				// シフトキー押下・押上時
				if (keyCode == VK_LSHIFT || keyCode == VK_RSHIFT)
				{
					if (isDown)
					{
						this.keyShift = true;
					}
					else
					{
						this.keyShift = false;
					}
				}
				// コントロールキー押下・押上時
				else if (keyCode == VK_LCONTROL || keyCode == VK_RCONTROL)
				{
					if (isDown)
					{
						this.keyCtrl = true;
					}
					else
					{
						this.keyCtrl = false;
					}
				}
				// Altキー押下・押上時
				else if (keyCode == VK_LSHIFT || keyCode == VK_RSHIFT)
				{
					if (isDown)
					{
						this.keyAlt = true;
					}
					else
					{
						this.keyAlt = false;
					}
				}
				// 通常キー押下・押上時
				else
				{
					if (e.isDown == false)
					{
						return;
					}
					keybordText = kc.ConvertToString(keyCode);
					isKeybordText = true;
				}

				//--------------------------
				// キーボードテキスト処理
				//--------------------------
				string modifiersText = "";
				if (this.keyShift)
				{
					modifiersText += "Shift ";
				}
				if (this.keyCtrl)
				{
					modifiersText += "Ctrl ";
				}
				if (this.keyAlt)
				{
					modifiersText += "Alt ";
				}

				this.lblKeyboard.Text = modifiersText + keybordText;
				if (isKeybordText)
				{
					addKeyHistory(modifiersText + keybordText);
				}
			}
		}
		#endregion 【イベント】 キーフック　キー入力イベント

		#region 【イベント】マウスフックイベント
		/// <summary>
		/// マウスイベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void Hook_MouseMoveEvent(object sender, MouseEventArg e)
		{
			//----------------------
			// マウス状態判定処理
			//----------------------
			// マウステキスト
			string mouseText = "";
			// マウス左ボタン押下・押上時
			if (e.mouseButton == MouseButton.Left)
			{
				if (e.isDown)
				{
					this.mouseLeft = true;
					this.pnlLeftButton.BackColor = this.colDown;
				}
				else
				{
					this.mouseLeft = false;
					this.pnlLeftButton.BackColor = this.colUp;
				}
			}
			// マウス右ボタン押下・押上時
			if (e.mouseButton == MouseButton.Right)
			{
				if (e.isDown)
				{
					this.mouseRight = true;
					this.pnlRightButton.BackColor = this.colDown;
				}
				else
				{
					this.mouseRight = false;
					this.pnlRightButton.BackColor = this.colUp;
				}
			}
			// マウス中ボタン押下・押上時
			if (e.mouseButton == MouseButton.Middle)
			{
				if (e.isDown)
				{
					this.mouseMiddle = true;
					this.pnlMiddleButton.BackColor = this.colDown;
				}
				else
				{
					this.mouseMiddle = false;
					this.pnlMiddleButton.BackColor = this.colUp;
				}
			}
			//----------------------
			// マウステキスト処理
			//----------------------
			if (this.mouseLeft)
			{
				mouseText += "Left ";
			}
			if (this.mouseMiddle)
			{
				mouseText += "Middle ";
			}
			if (this.mouseRight)
			{
				mouseText += "Right ";
			}
			this.lblMouseText.Text = mouseText;
		}
		#endregion 【イベント】マウスフックイベント

		#region 【イベント】フォームクロージングイベント
		/// <summary>
		/// フォームクロージングイベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			// 設定保存
			this.SaveSettings();

			// キーボードフック解放
			this.keyboardHook.UnHook();
			// マウスフック解放
			this.mouseHook.UnHook();

		}
		#endregion 【イベント】フォームクロージングイベント

		#region 【イベント】キー履歴非表示タイマ ティックイベント
		/// <summary>
		/// キー履歴非表示タイマ ティックイベント
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void timerHistoryEraser_Tick(object sender, EventArgs e)
		{
			removeKeyHistory();
			updateKeyHistoryText();
		}
		#endregion 【イベント】キー履歴非表示タイマ ティックイベント

		//----------------------------------
		// private メソッド
		//----------------------------------

		#region 【ロジック】キーボード入力履歴に追加
		/// <summary>
		/// キーボード入力履歴に追加
		/// </summary>
		/// <param name="keyText">追加する文字列</param>
		private void addKeyHistory(string keyText)
		{
			// 入力時刻を保持
			DateTime now = DateTime.Now;

			// Threadもどき待機処理
			while (this.isKeyHistroiesWorking)
			{
				Thread.Sleep(50);
			}

			// 処理中フラグ設定
			this.isKeyHistroiesWorking = true;

			// 履歴リストに追加
			this.keyHistroies.Add(now, keyText);

			// 処理中フラグ解除
			this.isKeyHistroiesWorking = false;
		}
		#endregion 【ロジック】キーボード入力履歴に追加

		#region 【ロジック】キーボード入力履歴から削除
		/// <summary>
		/// キーボード入力履歴から削除
		/// </summary>
		private void removeKeyHistory()
		{
			// Threadもどき待機処理
			while (this.isKeyHistroiesWorking)
			{
				Thread.Sleep(100);
			}

			// 処理中フラグ設定
			this.isKeyHistroiesWorking = true;

			// 削除対象時刻を取得
			DateTime pastTime = DateTime.Now.AddSeconds(-4);

			// 履歴数を一時変数に取得する(ループ内で.Countにアクセスしない)
			int len = keyHistroies.Count;

			// 履歴数分ループ
			for (int i = 0; i < len; i++)
			{
				// DateTimeキーを取得
				DateTime keyTime = keyHistroies.Keys[0];
				// 対象時刻以内なら
				if (keyTime > pastTime)
				{
					// 処理終了
					break;
				}
				// 履歴から1つ削除
				keyHistroies.RemoveAt(0);
			}

			// 処理中フラグ解除
			this.isKeyHistroiesWorking = false;
		}
		#endregion 【ロジック】キーボード入力履歴から削除

		#region 【ロジック】キーボード入力履歴の表示文字列更新
		/// <summary>
		/// キーボード入力履歴の表示文字列更新
		/// </summary>
		private void updateKeyHistoryText()
		{
			// Threadもどき待機処理
			while (this.isKeyHistroiesWorking)
			{
				Thread.Sleep(50);
			}

			// 処理中フラグ設定
			this.isKeyHistroiesWorking = true;

			// 履歴数を一時変数に取得する
			int len = keyHistroies.Count;
			int offset = 0;
			// 表示数の調整
			if (len > 5) offset = len - 5;      // 履歴5個以上は表示しない
			if (len == 1) len = 0;      // データ1個なら履歴表示なし

			// 履歴文字列を生成するための変数
			StringBuilder keyHistroiesText = new StringBuilder();
			// 履歴数分ループ
			for (int i = offset; i < len; i++)
			{
				// i番目の文字を取得
				string tmp = keyHistroies.Values[i];
				// 空文字なら次へ
				if (tmp == "") continue;
				// 履歴文字列に追加
				keyHistroiesText.AppendLine(tmp);
			}
			// ラベルにセット
			this.lblKeyHistory.Text = keyHistroiesText.ToString();

			// 履歴がすべてなくなったら
			if (keyHistroies.Count == 0)
			{
				// キーのメイン表示ラベルもブランクにする
				this.lblKeyboard.Text = "";
			}

			// 処理中フラグ解除
			this.isKeyHistroiesWorking = false;
		}
		#endregion 【ロジック】キーボード入力履歴の表示文字列更新

		#region 【ロジック】設定保存
		/// <summary>
		/// 設定保存
		/// </summary>
		private void SaveSettings()
		{
			SettingData obj = new SettingData();

			//------------------------
			// 設定データ
			//------------------------
			obj.formLocationX = this.Location.X;
			obj.formLocationY = this.Location.Y;

			//------------------------
			// XML保存
			//------------------------
			//string dir = System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
			string dir = Application.StartupPath;
			string fileName = System.IO.Path.Combine(dir, Application.ProductName + ".xml");

			System.IO.StreamWriter sw = null;
			try
			{
				//XmlSerializerオブジェクトを作成
				//オブジェクトの型を指定する
				System.Xml.Serialization.XmlSerializer serializer =
					new System.Xml.Serialization.XmlSerializer(typeof(SettingData));
				//書き込むファイルを開く（UTF-8 BOM無し）
				sw = new System.IO.StreamWriter(
					fileName, false, new System.Text.UTF8Encoding(false));
				//シリアル化し、XMLファイルに保存する
				serializer.Serialize(sw, obj);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message + ex.InnerException);
			}
			finally
			{
				if (sw != null)
				{
					//ファイルを閉じる
					sw.Close();
				}
			}
		}
		#endregion 【ロジック】設定保存

		#region 【ロジック】設定読込
		/// <summary>
		/// 設定読込
		/// </summary>
		private void LoadSettings()
		{
			//保存元のファイル名
			//string dir = System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
			string dir = Application.StartupPath;
			string fileName = System.IO.Path.Combine(dir, Application.ProductName + ".xml");

			//XmlSerializerオブジェクトを作成
			SettingData obj = null;
			System.IO.StreamReader sr = null;
			try
			{
				// 設定ファイルがまだなければ読み込み処理なし
				if (File.Exists(fileName) == false)
				{
					return;
				}

				System.Xml.Serialization.XmlSerializer serializer
					= new System.Xml.Serialization.XmlSerializer(typeof(SettingData));
				//読み込むファイルを開く
				sr = new System.IO.StreamReader(
					fileName, new System.Text.UTF8Encoding(false));
				//XMLファイルから読み込み、逆シリアル化する
				obj = serializer.Deserialize(sr) as SettingData;

				//------------------------
				// 設定データ
				//------------------------
				this.Location = new Point(obj.formLocationX, obj.formLocationY);
			}
			catch
			{
			}
			finally
			{
				if (sr != null)
				{
					//ファイルを閉じる
					sr.Close();
				}
			}
		}
		#endregion 【ロジック】設定読込


		//----------------------------------
		// メニューイベント
		//----------------------------------

		#region 【メニュー】ウィンドウの枠表示・非表示切り替え
		/// <summary>
		/// ウィンドウの枠表示・非表示切り替え
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void showFrameToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (this.FormBorderStyle != FormBorderStyle.None)
			{
				this.FormBorderStyle = FormBorderStyle.None;
				this.pnlFrame.BackColor = this.TransparencyKey;     // パネルをForm透明色にする
				this.pnlInductionFrame.Visible = false;                 // 右クリックできます 非表示
			}
			else
			{
				this.FormBorderStyle = FormBorderStyle.FixedSingle;
				this.pnlFrame.BackColor = Color.PaleTurquoise;     // パネルをForm非透明色にする
				this.pnlInductionFrame.Visible = true;                  // 右クリックできます 表示
			}

			// 設定保存
			this.SaveSettings();
		}
		#endregion 【メニュー】ウィンドウの枠表示・非表示切り替え

		#region 【メニュー】閉じるメニュー
		/// <summary>
		/// 閉じるメニュー
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void closeToolStripMenuItem_Click(object sender, EventArgs e)
		{
			this.Close();
		}
		#endregion 【メニュー】閉じるメニュー

	}
}
