﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.Drawing;

namespace InputViewer
{
	/// <summary>
	/// マウスフック クラス
	/// </summary>
	internal class MouseHook
	{
		#region 【Win32　定数・構造体】
		protected const int WH_MOUSE_LL = 14;

		// ウィンドウズメッセージ定義
		protected const int WM_MOUSEMOVE = 0x0200;      // マウス移動
		protected const int WM_LBUTTONDOWN = 0x0201;    // マウス左ボタン ダウン
		protected const int WM_LBUTTONUP = 0x0202;      // マウス左ボタン アップ
		protected const int WM_RBUTTONDOWN = 0x0204;    // マウス右ボタン ダウン
		protected const int WM_RBUTTONUP = 0x0205;      // マウス右ボタン アップ
		protected const int WM_MBUTTONDOWN = 0x0207;    // マウス中ボタン ダウン
		protected const int WM_MBUTTONUP = 0x0208;      // マウス中ボタン アップ
		protected const int WM_MOUSEWHEEL = 0x020A;     // マウスホイール

		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr SetWindowsHookEx(int idHook, MouseHookProc lpfn, IntPtr hMod, uint dwThreadId);

		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool UnhookWindowsHookEx(IntPtr hhk);

		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr CallNextHookEx(IntPtr hhk, int nCode, IntPtr wParam, IntPtr lParam);

		[DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr GetModuleHandle(string lpModuleName);

		// これをしないとGCにより回収されてしまってCallbackOnCollectedDelegate例外で詰む
		private delegate IntPtr MouseHookProc(int nCode, IntPtr wParam, IntPtr lParam);
		MouseHookProc mouseHookProc = null;

		private IntPtr hookId = IntPtr.Zero;

		[StructLayout(LayoutKind.Sequential)]
		private struct POINT
		{
			public int x;
			public int y;
		}

		[StructLayout(LayoutKind.Sequential)]
		private struct MSLLHOOKSTRUCT
		{
			public POINT pt;
			public uint mouseData;
			public uint flags;
			public uint time;
			public System.IntPtr dwExtraInfo;
		}
		#endregion 【Win32　定数・構造体】

		public delegate void MouseEventHandler(object sender, MouseEventArg e);
		public event MouseEventHandler MouseEvent;

		#region 【メソッド】マウス フック
		/// <summary>
		/// マウス フック
		/// </summary>
		public void Hook()
		{
			if (hookId == IntPtr.Zero)
			{
				using (var curProcess = Process.GetCurrentProcess())
				{
					using (ProcessModule curModule = curProcess.MainModule)
					{
						// フィールド変数にHookProcedureを保存する
						// そうしないとGCにより回収されてしまってCallbackOnCollectedDelegate例外で詰む
						mouseHookProc = HookProcedure;
						hookId = SetWindowsHookEx(WH_MOUSE_LL, mouseHookProc, GetModuleHandle(curModule.ModuleName), 0);
					}
				}
			}
		}
		#endregion 【メソッド】マウス フック

		#region 【メソッド】マウス アンフック
		/// <summary>
		/// アンフック
		/// </summary>
		public void UnHook()
		{
			UnhookWindowsHookEx(hookId);
			hookId = IntPtr.Zero;
		}
		#endregion 【メソッド】マウス アンフック

		#region 【メソッド】マウスフック プロシージャ
		/// <summary>
		/// マウスフック　プロシージャ
		/// </summary>
		/// <param name="nCode"></param>
		/// <param name="wParam"></param>
		/// <param name="lParam"></param>
		/// <returns></returns>
		public IntPtr HookProcedure(int nCode, IntPtr wParam, IntPtr lParam)
		{
			// マウス移動
			if (nCode >= 0 && wParam == (IntPtr)WM_MOUSEMOVE)
			{
				MSLLHOOKSTRUCT ms = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
				// Invoke からメインスレッド処理呼び出し
				MouseEvent?.Invoke(this, makeMouseEventArg(ms, MouseButton.Move, true));
			}
			// 左ボタン　ダウン
			if (nCode >= 0 && wParam == (IntPtr)WM_LBUTTONDOWN)
			{
				MSLLHOOKSTRUCT ms = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
				// Invoke からメインスレッド処理呼び出し
				MouseEvent?.Invoke(this, makeMouseEventArg(ms, MouseButton.Left, true));
			}
			// 左ボタン　アップ
			if (nCode >= 0 && wParam == (IntPtr)WM_LBUTTONUP)
			{
				MSLLHOOKSTRUCT ms = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
				// Invoke からメインスレッド処理呼び出し
				MouseEvent?.Invoke(this, makeMouseEventArg(ms, MouseButton.Left, false));
			}
			// 右ボタン　ダウン
			if (nCode >= 0 && wParam == (IntPtr)WM_RBUTTONDOWN)
			{
				MSLLHOOKSTRUCT ms = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
				// Invoke からメインスレッド処理呼び出し
				MouseEvent?.Invoke(this, makeMouseEventArg(ms, MouseButton.Right, true));
			}
			// 右ボタン　アップ
			if (nCode >= 0 && wParam == (IntPtr)WM_RBUTTONUP)
			{
				MSLLHOOKSTRUCT ms = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
				// Invoke からメインスレッド処理呼び出し
				MouseEvent?.Invoke(this, makeMouseEventArg(ms, MouseButton.Right, false));
			}
			// 中ボタン　ダウン
			if (nCode >= 0 && wParam == (IntPtr)WM_MBUTTONDOWN)
			{
				MSLLHOOKSTRUCT ms = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
				// Invoke からメインスレッド処理呼び出し
				MouseEvent?.Invoke(this, makeMouseEventArg(ms, MouseButton.Middle, true));
			}
			// 中ボタン　アップ
			if (nCode >= 0 && wParam == (IntPtr)WM_MBUTTONUP)
			{
				MSLLHOOKSTRUCT ms = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
				// Invoke からメインスレッド処理呼び出し
				MouseEvent?.Invoke(this, makeMouseEventArg(ms, MouseButton.Middle, false));
			}
			// マウスホイール
			if (nCode >= 0 && wParam == (IntPtr)WM_MOUSEWHEEL)
			{
				MSLLHOOKSTRUCT ms = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
				// Invoke からメインスレッド処理呼び出し
				MouseEvent?.Invoke(this, makeMouseEventArg(ms, MouseButton.Wheel, true));
			}

			return CallNextHookEx(hookId, nCode, wParam, lParam);
		}
		#endregion 【メソッド】マウスフック プロシージャ

		#region 【メソッド】変換 MSLLHOOKSTRUCT構造体→MouseEventArgクラス
		private MouseEventArg makeMouseEventArg(MSLLHOOKSTRUCT ms, MouseButton a_mouseButton, bool a_isDown)
		{
			return new MouseEventArg(ms.pt.x, ms.pt.y, ms.mouseData, ms.flags, ms.time, a_mouseButton, a_isDown);
		}
		#endregion 【メソッド】変換 MSLLHOOKSTRUCT構造体→MouseEventArgクラス

	}

	/// <summary>
	/// マウスボタン列挙体
	/// </summary>
	internal enum MouseButton
	{
		None,
		Move,
		Left,
		Right,
		Middle,
		Wheel
	}

	/// <summary>
	/// マウスイベントクラス
	/// </summary>
	internal class MouseEventArg : EventArgs
	{
		// 処理高速化のためgetter/setter は不要
		public Point point;
		public uint mouseData;
		public uint flags;
		public uint time;

		public bool isDown;
		public MouseButton mouseButton = MouseButton.None;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="a_x"></param>
		/// <param name="a_y"></param>
		/// <param name="a_mouseData"></param>
		/// <param name="a_flags"></param>
		/// <param name="a_time"></param>
		/// <param name="a_mouseButton"></param>
		/// <param name="a_isDown"></param>
		public MouseEventArg(int a_x, int a_y, uint a_mouseData, uint a_flags, uint a_time, MouseButton a_mouseButton, bool a_isDown)
		{
			this.point = new Point(a_x, a_y);
			this.mouseData = a_mouseData;
			this.flags = a_flags;
			this.time = a_time;
			this.isDown = a_isDown;
			this.mouseButton = a_mouseButton;
		}
	}
}
